/*********************************************************************************/
/*  Copyright 2015 Barcelona Supercomputing Center                               */
/*                                                                               */
/*  This file is part of the DLB library.                                        */
/*                                                                               */
/*  DLB is free software: you can redistribute it and/or modify                  */
/*  it under the terms of the GNU Lesser General Public License as published by  */
/*  the Free Software Foundation, either version 3 of the License, or            */
/*  (at your option) any later version.                                          */
/*                                                                               */
/*  DLB is distributed in the hope that it will be useful,                       */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of               */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                */
/*  GNU Lesser General Public License for more details.                          */
/*                                                                               */
/*  You should have received a copy of the GNU Lesser General Public License     */
/*  along with DLB.  If not, see <http://www.gnu.org/licenses/>.                 */
/*********************************************************************************/

#define _GNU_SOURCE
#include <sched.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/resource.h>

#include "LB_comm/shmem.h"
#include "LB_comm/shmem_stats.h"
#include "LB_numThreads/numThreads.h"
#include "support/debug.h"
#include "support/mask_utils.h"
#include "support/mytime.h"

#define NOBODY 0
#define ME getpid()
static const long UPDATE_USAGE_MIN_THRESHOLD    =  100000000L;   // 10^8 ns = 100ms
//static const long UPDATE_LOADAVG_MIN_THRESHOLD  = 1000000000L;   // 10^9 ns = 1s
//static const double LOG2E = 1.44269504088896340736;

typedef pid_t spid_t;  // Sub-process ID

typedef struct {
    spid_t pid;
    unsigned int active_cpus;
    // Cpu Usage fields:
    double cpu_usage;
    double cpu_avg_usage;
    struct timespec last_ttime; // Total time
    struct timespec last_utime; // Useful time (user+system)
#ifdef DLB_LOAD_AVERAGE
    // Load average fields:
    float load[3];              // 1min, 5min, 15mins
    struct timespec last_ltime; // Last time that Load was updated
#endif
} pinfo_t;

typedef struct {
    stats_state_t current_state;
    struct timespec last_update;
    int64_t acc_time[3];        // Accumulated time for each state
} cpuinfo_t;

/* We cannot have two different structures of 0 length in the shared memory.
 * Although, like both are of the same size, they can be indexed with the same array
 */
typedef struct {
    pinfo_t     process;
    cpuinfo_t   cpu;
} info_t;

typedef struct {
    struct timespec initial_time;
    info_t info[0];             // info can be indexed by either process or CPU id,
                                // but you should access only the corresponding field
} shdata_t;

// Since accessing both processes and CPUs through the same structure and index may
// affect the code readability, these macros may be useful
#define shdata_process(i) shdata->info[(i)].process
#define shdata_cpu(i) shdata->info[(i)].cpu

static shmem_handler_t *shm_handler = NULL;
static shmem_handler_t *shm_ext_handler = NULL;
static shdata_t *shdata = NULL;
static int max_processes;
static int my_process;

/*** Helper functions, the shm lock may have to be acquired beforehand ***/
static float getcpustate(int cpu, stats_state_t state) {
    struct timespec now;
    get_time_coarse( &now );
    int64_t total = timespec_diff(&shdata->initial_time, &now);
    int64_t acc = shdata_cpu(cpu).acc_time[state];
    float percentage = (float)acc/total;
    ensure( percentage >= -0.000001f && percentage <= 1.000001f,
                "Percentage out of bounds" );
    return percentage;
}


static void print_stats(void) {
    fprintf(stdout, "=== DLB Statistics ===\n");
    fprintf(stdout, "Process %d: %f CPU Avg usage\n",
            my_process, shdata_process(my_process).cpu_avg_usage);

    fprintf(stdout, "CPU Statistics:\n");
    int i;
    for (i=0; i<max_processes; ++i) {
        fprintf(stdout, "CPU %d: OWNED(%.2f%%), GUESTED(%.2f%%), IDLE(%.2f%%)\n",
                i,
                getcpustate(i, STATS_OWNED)*100,
                getcpustate(i, STATS_GUESTED)*100,
                getcpustate(i, STATS_IDLE)*100);
    }
}
/*** End of helper functions ***/


void shmem_stats__init( void ) {
    // Protect double initialization
    if ( shm_handler != NULL ) {
        warning("DLB_Stats is being initialized more than once");
        return;
    }

    // We will reserve enough positions assuming that there won't be more processes than cpus
    max_processes = mu_get_system_size();

    // Basic size + zero-length array real length
    shmem_handler_t *init_handler = shmem_init( (void**)&shdata,
            sizeof(shdata_t) + sizeof(info_t)*max_processes, "stats" );

    shmem_lock( init_handler );
    {
        int p;
        for ( p = 0; p < max_processes; p++ ) {
            if ( shdata_process(p).pid == NOBODY ) {
                shdata_process(p).pid = ME;
#ifdef DLB_LOAD_AVERAGE
                shdata_process(p).load[0] = 0.0f;
                shdata_process(p).load[1] = 0.0f;
                shdata_process(p).load[2] = 0.0f;
#endif
                my_process = p;
                break;
            }
        }
        fatal_cond( p == max_processes, "Cannot reserve process statistics" );

        // Initialize initial time and CPU timing on shmem creation
        if ( shdata->initial_time.tv_sec == 0 && shdata->initial_time.tv_nsec == 0 ) {
            get_time( &shdata->initial_time );

            for ( p = 0; p < max_processes; p++ ) {
                shdata_cpu(p).last_update = shdata->initial_time;
            }
        }
    }
    shmem_unlock( init_handler );

    // Global variable is only assigned after the initialization
    shm_handler = init_handler;
}

void shmem_stats__finalize( void ) {
    shmem_lock( shm_handler );
    {
        print_stats();
        shdata_process(my_process).pid = NOBODY;
    }
    shmem_unlock( shm_handler );
    shmem_finalize( shm_handler );
    shm_handler = NULL;
}

void shmem_stats__update( void ) {

    if (shm_handler == NULL || shdata == NULL) return;

    // Do not update if the elapsed total time is less than a predefined threshold
    struct timespec now;
    int64_t elapsed;
    get_time_coarse( &now );
    elapsed = timespec_diff( &shdata_process(my_process).last_ttime, &now );
    if ( elapsed < UPDATE_USAGE_MIN_THRESHOLD )
        return;

    shmem_lock( shm_handler );
    {
        // Get the active CPUs
        cpu_set_t mask;
        get_mask( &mask );
        shdata_process(my_process).active_cpus = CPU_COUNT( &mask );

        // Compute elapsed total time
        struct timespec current_ttime;
        int64_t elapsed_ttime_since_last;
        int64_t elapsed_ttime_since_init;;
        get_time_coarse( &current_ttime );
        elapsed_ttime_since_last = timespec_diff(&shdata_process(my_process).last_ttime,
                &current_ttime);
        elapsed_ttime_since_init = timespec_diff(&shdata->initial_time, &current_ttime );

        // Compute elapsed useful time (user+system)
        struct rusage usage;
        struct timespec current_utime;
        int64_t elapsed_utime_since_last;
        int64_t elapsed_utime_since_init;
        getrusage( RUSAGE_SELF, &usage );
        add_tv_to_ts( &usage.ru_utime, &usage.ru_stime, &current_utime );
        elapsed_utime_since_last = timespec_diff(&shdata_process(my_process).last_utime,
                &current_utime );
        elapsed_utime_since_init = ts_to_ns(&current_utime);

        // Update times for next update
        shdata_process(my_process).last_ttime = current_ttime;
        shdata_process(my_process).last_utime = current_utime;

        // Compute usage
        shdata_process(my_process).cpu_usage = 100 *
            (double)elapsed_utime_since_last / (double)elapsed_ttime_since_last;

        // Compute avg usage
        shdata_process(my_process).cpu_avg_usage = 100 *
            (double)elapsed_utime_since_init / (double)elapsed_ttime_since_init;

#ifdef DLB_LOAD_AVERAGE
        // Do not update the Load Average if the elapsed is less that a threshold
        if ( elapsed_ttime > UPDATE_LOADAVG_MIN_THRESHOLD ) {
            shdata_process(my_process).last_ltime = current_ttime;
            // WIP
        }
#endif
    }
    shmem_unlock( shm_handler );
}

void shmem_stats__update_cpu(int cpu, stats_state_t new_state) {

    if (shm_handler == NULL || shdata == NULL) return;

    shmem_lock( shm_handler );
    {
        struct timespec now;
        int64_t elapsed;

        // Compute elapsed since last update
        get_time( &now );
        elapsed = timespec_diff( &shdata_cpu(cpu).last_update, &now );

        // Update fields
        stats_state_t current_state = shdata_cpu(cpu).current_state;
        shdata_cpu(cpu).acc_time[current_state] += elapsed;
        shdata_cpu(cpu).current_state = new_state;
        shdata_cpu(cpu).last_update = now;
    }
    shmem_unlock( shm_handler );
}


/* From here: functions aimed to be called from an external process, only to consult the shmem */

void shmem_stats_ext__init( void ) {
    // Protect multiple initialization
    if ( shm_ext_handler != NULL ) {
        warning("DLB_Stats is being initialized more than once");
        return;
    }

    max_processes = mu_get_system_size();
    shm_ext_handler = shmem_init( (void**)&shdata, sizeof(shdata_t) + sizeof(info_t)*max_processes, "stats" );
}

void shmem_stats_ext__finalize( void ) {
    shmem_finalize( shm_ext_handler );
    shm_ext_handler = NULL;
}

int shmem_stats_ext__getnumcpus( void ) {
    return mu_get_system_size();
}

void shmem_stats_ext__getpidlist( int *pidlist, int *nelems, int max_len ) {
    *nelems = 0;
    if (shm_ext_handler == NULL) return;
    shmem_lock( shm_ext_handler );
    {
        int p;
        for ( p = 0; p < max_processes; p++ ) {
            int pid = shdata_process(p).pid;
            if ( pid != NOBODY ) {
                pidlist[(*nelems)++] = pid;
            }
            if ( *nelems == max_len ) {
                break;
            }
        }
    }
    shmem_unlock( shm_ext_handler );
}

double shmem_stats_ext__getcpuusage( int pid ) {
    if (shm_ext_handler == NULL) return -1.0;

    double cpu_usage = -1.0;
    shmem_lock( shm_ext_handler );
    {
        int p;
        for ( p = 0; p < max_processes; p++ ) {
            if ( shdata_process(p).pid == pid ) {
                cpu_usage = shdata_process(p).cpu_usage;
                break;
            }
        }
    }
    shmem_unlock( shm_ext_handler );

    return cpu_usage;
}

double shmem_stats_ext__getcpuavgusage( int pid ) {
    if (shm_ext_handler == NULL) return -1.0;

    double cpu_avg_usage = -1.0;
    shmem_lock( shm_ext_handler );
    {
        int p;
        for ( p = 0; p < max_processes; p++ ) {
            if ( shdata_process(p).pid == pid ) {
                cpu_avg_usage = shdata_process(p).cpu_avg_usage;
                break;
            }
        }
    }
    shmem_unlock( shm_ext_handler );

    return cpu_avg_usage;
}

void shmem_stats_ext__getcpuusage_list( double *usagelist, int *nelems, int max_len ) {
    *nelems = 0;
    if (shm_ext_handler == NULL) return;
    shmem_lock( shm_ext_handler );
    {
        int p;
        for ( p = 0; p < max_processes; p++ ) {
            if ( shdata_process(p).pid != NOBODY ) {
                usagelist[(*nelems)++] = shdata_process(p).cpu_usage;
            }
            if ( *nelems == max_len ) {
                break;
            }
        }
    }
    shmem_unlock( shm_ext_handler );
}

void shmem_stats_ext__getcpuavgusage_list( double *avgusagelist, int *nelems, int max_len ) {
    *nelems = 0;
    if (shm_ext_handler == NULL) return;
    shmem_lock( shm_ext_handler );
    {
        int p;
        for ( p = 0; p < max_processes; p++ ) {
            if ( shdata_process(p).pid != NOBODY ) {
                avgusagelist[(*nelems)++] = shdata_process(p).cpu_avg_usage;
            }
            if ( *nelems == max_len ) {
                break;
            }
        }
    }
    shmem_unlock( shm_ext_handler );
}

double shmem_stats_ext__getnodeusage(void) {
    if (shm_ext_handler == NULL) return -1.0;

    double cpu_usage = 0.0;
    shmem_lock( shm_ext_handler );
    {
        int p;
        for ( p = 0; p < max_processes; p++ ) {
            if ( shdata_process(p).pid != NOBODY ) {
                cpu_usage += shdata_process(p).cpu_usage;
            }
        }
    }
    shmem_unlock( shm_ext_handler );

    return cpu_usage;
}

double shmem_stats_ext__getnodeavgusage(void) {
    if (shm_ext_handler == NULL) return -1.0;

    double cpu_avg_usage = 0.0;
    shmem_lock( shm_ext_handler );
    {
        int p;
        for ( p = 0; p < max_processes; p++ ) {
            if ( shdata_process(p).pid != NOBODY ) {
                cpu_avg_usage += shdata_process(p).cpu_avg_usage;
            }
        }
    }
    shmem_unlock( shm_ext_handler );

    return cpu_avg_usage;
}

int shmem_stats_ext__getactivecpus( int pid ) {
    if (shm_ext_handler == NULL) return -1;

    int active_cpus = -1;
    shmem_lock( shm_ext_handler );
    {
        int p;
        for ( p = 0; p < max_processes; p++ ) {
            if ( shdata_process(p).pid == pid ) {
                active_cpus = shdata_process(p).active_cpus;
                break;
            }
        }
    }
    shmem_unlock( shm_ext_handler );
    return active_cpus;
}

void shmem_stats_ext__getactivecpus_list( int *cpuslist, int *nelems, int max_len ) {
    *nelems = 0;
    if (shm_ext_handler == NULL) return;
    shmem_lock( shm_ext_handler );
    {
        int p;
        for ( p = 0; p < max_processes; p++ ) {
            if ( shdata_process(p).pid != NOBODY ) {
                cpuslist[(*nelems)++] = shdata_process(p).active_cpus;
            }
            if ( *nelems == max_len ) {
                break;
            }
        }
    }
    shmem_unlock( shm_ext_handler );
}

int shmem_stats_ext__getloadavg( int pid, double *load ) {
    if (shm_ext_handler == NULL) return -1;
    int error = -1;
#ifdef DLB_LOAD_AVERAGE
    shmem_lock( shm_ext_handler );
    {
        int p;
        for ( p = 0; p < max_processes; p++ ) {
            if ( shdata_process(p).pid == pid ) {
                load[0] = shdata_process(p).load[0];
                load[1] = shdata_process(p).load[1];
                load[2] = shdata_process(p).load[2];
                error = 0;
                break;
            }
        }
    }
    shmem_unlock( shm_ext_handler );
#endif
    return error;
}

float shmem_stats_ext__getcpustate(int cpu, stats_state_t state) {
    return getcpustate(cpu, state);
}

void shmem_stats_ext__print_info(void) {
    if (shm_ext_handler == NULL) return;
    shmem_lock( shm_ext_handler );
    {
        print_stats();
    }
    shmem_unlock( shm_ext_handler );
}
